import express from 'express';
import got from 'got';
import { inspect } from 'util';

const app = express();
const port = 4000;
const cache = {};

// Wrapped in a function to avoid polluting global namespace with
// temporary variables.
const config = await (async function () {
    let config = (await import('./config.js')).default;
    let overrides;
    try {
        overrides = (await import('./config.override.js')).default;
    }
    catch (e) {
        if (e.code == null || e.code != 'ERR_MODULE_NOT_FOUND') {
            throw e;
        }

        overrides = {};
    }

    let composed = { ...config, ...overrides };
    console.log('Config:', composed);
    return composed;
})();

let token = null;

const api = got.extend({
    prefixUrl: config.url,
    throwHttpErrors: false,
    retry: { limit: 0 },
    responseType: 'json',
    hooks: {
        beforeRequest: [
            options => {
                if (token != null) {
                    options.headers['Authorization'] =
                        'Bearer ' + token.access_token;
                }
            }
        ],
        afterResponse: [
            response => {
                // Print requests to the console
                console.debug(
                    response.statusCode,
                    response.request.options.method,
                    response.requestUrl.href
                );

                return response;
            }
        ]
    },
    pagination: {
        // Got doesn't support relative URLs.
        // Overridden to provide basic support.
        paginate: ({ response }) => {
            let linkHeader = response.headers.link;
            let next = null;
            for (let link of linkHeader.split(',').map(h => h.trim())) {
                let match = link.match(/<(?<link>.+)>;\s+rel="next"/i);
                if (match != null) {
                    next = match.groups.link;
                }
            }

            if (next == null) {
                return false;
            }

            if (next.startsWith('/')) {
                next = next.substring(1);
            }

            return {
                url: new URL(next, new URL(config.url).origin)
            };
        }
    },
    https: {
        rejectUnauthorized: false
    }
});

function checkAndLogError(apiRes, proxyRes) {
    if (apiRes.statusCode >= 200 && apiRes.statusCode < 300) {
        return true;
    }

    let data = {
        statusCode: apiRes.statusCode,
        headers: apiRes.headers,
        body: apiRes.body
    };
    console.error(
        'API request failed.',
        inspect(data, true, null, true)
    );
    proxyRes
        .status(apiRes.statusCode == 401 ? 401 : 500)
        .send({ error: apiRes.body });
    return false;
}

app.use(express.static('public'));
app.use(express.json());
app.disable('etag');

/**
 * Authentication
 */

app.get('/authenticate', async (req, res) => {
    let result = await api.post('authenticate', {
        json: {
            api_key: config.apiKey,
            client_db: config.clientDb,
            username: config.username
        }
    });

    if (!checkAndLogError(result, res)) {
        return;
    }

    token = result.body;
    res.send(token);
});

app.get('/is-authenticated', (req, res) => {
    res.json(token != null);
});

app.get('/token', (req, res) => {
    if (token == null) {
        res.status(404).send();
        return;
    }

    res.json(token);
});

app.delete('/token', (req, res) => {
    token = null;
    res.send();
});

/**
 *  Data management
 */

function createCachedProxy(localRoute, cacheKey, remoteRoute) {
    app.get(localRoute, async (req, res) => {
        if (cache[cacheKey]) {
            res.json(cache[cacheKey]);
            return;
        }

        // Get up to 100 results.
        let response = await api.paginate.all(remoteRoute, {
            pagination: {
                countLimit: 100
            }
        });

        cache[cacheKey] = response;
        res.json(response);
    });
}

createCachedProxy(
    '/service_request_types',
    'service_request_types',
    'maintenance/resources/service_request_types'
);

createCachedProxy(
    '/customers',
    'customers',
    'maintenance/resources/customers'
);

createCachedProxy(
    '/properties',
    'properties',
    'maintenance/resources/properties'
);

createCachedProxy(
    '/departments',
    'departments',
    'classifications/departments'
);

createCachedProxy(
    '/work_orders',
    'work_orders',
    'maintenance/work_orders'
);

/**
 * Form submissions
 */

app.post('/service_request', async (req, res) => {
    let response = await api.post('maintenance/service_requests', {
        json: {
            service_request_type_id: req.body.typeId,
            status: req.body.status,
            priority: req.body.priority,
            customer_id: req.body.customer,
            property_id: req.body.property,
            department_id: req.body.department
        }
    });

    if (!checkAndLogError(response, res)) {
        return;
    }

    let location = response.headers.location || '';
    const createdId = location.substring(location.lastIndexOf('/') + 1);

    // Creating a Service Request results in a transaction because mandatory
    // attributes must be set up before it can be made available in the system.
    let txId = response.body.TransactionID;
    response = await api.post(`transactions/${txId}/commit`);
    if (!checkAndLogError(response, res)) {
        return;
    }

    res.send(createdId);
});

app.post('/attach_cost', async (req, res) => {
    let response = await api.post(
        `maintenance/work_orders/${req.body.workOrder}/costs`,
        {
            json: {
                description: req.body.description,
                units: req.body.units,
                cost_unit: req.body.costUnit,
                cost_date: new Date().toISOString().split('T')[0],
                // API only supports `Other` at the moment.
                source: "Other"
            }
        });

    if (!checkAndLogError(response, res)) {
        return;
    }

    let location = response.headers.location || '';
    res.send(location.substring(location.lastIndexOf('/') + 1));
});

/**
 *  Cache
 */

app.delete('/cache', (req, res) => {
    for (let key of Object.keys(cache)) {
        delete cache[key];
    }

    res.send();
});

app.get('/cache/keys', (req, res) => {
    res.json(Object.keys(cache));
});

/**
 *  Launch!
 */

app.listen(port, () => {
    console.log(`Example app listening on port ${port}`)
});
